/*
 *	Real Time Clock interface for Road MR-2A on-chip RTC
 *
 *	Based on v1.10 of Paul Gortmaker's PC RTC driver
 *	VR41xx adaptation by Michael Klar
 *	MR-2A adaptation by Jalen Yang
 *
 *	Copyright (C) 1996 Paul Gortmaker
 *	Copyright (C) 2000 Michael Klar
 *	Copyright (C) 2001 Jalen Yang
 *
 *	This driver allows use of the real time clock (built into
 *	nearly all computers) from user space. It exports the /dev/mr2artc
 *	interface supporting various ioctl() and also the
 *	/proc/driver/mr2artc pseudo-file for status information.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Based on other minimal char device drivers, like Alan's
 *	watchdog, Ted's random, etc. etc.
 */
/*        
 *      ChangeLog :
 *         Jackal modify : 24 hours -> 12 hours (PM/AM)
 */

#define MR2A_RTC_VERSION	"2003-03-10 V1.0"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>

#include <linux/delay.h> /* jackal add */

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/vr41xx.h>


extern spinlock_t rtc_lock;

#if 0 /* kernel version : 2.4.10 */
extern void set_rtc_time(unsigned long timeval, volatile unsigned short *reg);
#else /* kernel version : 2.4.18 */
extern int  rtc_set_time(unsigned long sec);
#endif

//static struct fasync_struct *rtc_async_queue;

static DECLARE_WAIT_QUEUE_HEAD(rtc_wait);

static int mr2a_rtc_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data);
static int set_mr2a_rtc_time(struct rtc_time *rtc_tm);

static void get_rtc_time(struct rtc_time *rtc_tm);

#if 1
unsigned long epoch_adj;
#endif
/*
 *	Bits in rtc_status. (6 bits of room for future expansion)
 */

#define RTC_IS_OPEN		0x01	/* means /dev/mr2artc is in use	*/

#define MR2A_RTC_SEC1	__preg16(KSEG1 + 0x0A000200)
#define MR2A_RTC_SEC10	__preg16(KSEG1 + 0x0A000202)
#define MR2A_RTC_MIN1	__preg16(KSEG1 + 0x0A000204)
#define MR2A_RTC_MIN10	__preg16(KSEG1 + 0x0A000206)
#define MR2A_RTC_HOU1	__preg16(KSEG1 + 0x0A000208)
#define MR2A_RTC_HOU10	__preg16(KSEG1 + 0x0A00020A)
#define MR2A_RTC_DAY1	__preg16(KSEG1 + 0x0A00020C)
#define MR2A_RTC_DAY10	__preg16(KSEG1 + 0x0A00020E)
#define MR2A_RTC_MON1	__preg16(KSEG1 + 0x0A000210)
#define MR2A_RTC_MON10	__preg16(KSEG1 + 0x0A000212)
#define MR2A_RTC_YEA1	__preg16(KSEG1 + 0x0A000214)
#define MR2A_RTC_YEA10	__preg16(KSEG1 + 0x0A000216)
#define MR2A_RTC_WEE	__preg16(KSEG1 + 0x0A000218)
#define MR2A_RTC_REGD	__preg16(KSEG1 + 0x0A00021A)
#define MR2A_RTC_REGE	__preg16(KSEG1 + 0x0A00021C)
#define MR2A_RTC_REGF	__preg16(KSEG1 + 0x0A00021E)

static atomic_t rtc_status = ATOMIC_INIT(0); /* bitmapped status byte.	*/
static unsigned long rtc_irq_data = 0;	/* our output to the world	*/


/*
 *	If this driver ever becomes modularised, it will be really nice
 *	to make the epoch retain its value across module reload...
 */

unsigned long epoch = 1970;	/* year corresponding to time 0		*/

extern unsigned long epoch_adj;	/* cached adjustment value, in seconds	*/

static const unsigned char days_in_mo[] = 
{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};


static void mr2a_rtc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
}


/*
 *	Now all the various file operations that we export.
 */

static long long mr2a_rtc_llseek(struct file *file, loff_t offset, int origin)
{
	return 0;
}

static ssize_t mr2a_rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long data;
	ssize_t retval;

	if (count < sizeof(unsigned long))
		return -EINVAL;

	add_wait_queue(&rtc_wait, &wait);

	current->state = TASK_INTERRUPTIBLE;
		
	while ((data = xchg(&rtc_irq_data, 0)) == 0) {
		if (file->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			goto out;
		}
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			goto out;
		}
		schedule();
	}

	retval = put_user(data, (unsigned long *)buf); 
	if (!retval)
		retval = sizeof(unsigned long); 
out:
	current->state = TASK_RUNNING;
	remove_wait_queue(&rtc_wait, &wait);

	return retval;
}

static int mr2a_rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
	struct rtc_time wtime;
   
	switch (cmd) {
	case RTC_RD_TIME:	/* Read the time/date from RTC	*/
	{
	   get_rtc_time(&wtime);
	   break;
	}
	case RTC_SET_TIME:	/* Set the RTC */
	{
	   struct rtc_time mr2a_rtc_tm;
	   
	   if (!capable(CAP_SYS_TIME))
	     return -EACCES;
	   
	   if (copy_from_user(&mr2a_rtc_tm, (struct rtc_time*)arg,
			      sizeof(struct rtc_time)))
	     return -EFAULT;
	   
	   return set_mr2a_rtc_time(&mr2a_rtc_tm);
	   
	}
	 case RTC_EPOCH_READ:	/* Read the epoch.	*/
	     {
		return put_user (epoch, (unsigned long *)arg);
	}
	case RTC_EPOCH_SET:	/* Set the epoch.	*/
	{
		/* 
		 * There were no RTC clocks before 2000.
		 */
		if (arg < 1970)
			return -EINVAL;

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		epoch = arg;
		epoch_adj = mktime(epoch,0,0,0,0,0) - mktime(1970,0,0,0,0,0);
		return 0;
	}
	default:
		return -EINVAL;
	}
	return copy_to_user((void *)arg, &wtime, sizeof wtime) ? -EFAULT : 0;
}

/*
 *	We enforce only one user at a time here with the open/close.
 *	Also clear the previous interrupt data on an open, and clean
 *	up things on a close.
 */

static int mr2a_rtc_open(struct inode *inode, struct file *file)
{
   
        if(atomic_read(&rtc_status) & RTC_IS_OPEN)
        return -EBUSY;
   
	MOD_INC_USE_COUNT;

	atomic_set(&rtc_status, atomic_read(&rtc_status) | RTC_IS_OPEN);

	spin_lock_irq(&rtc_lock);
	rtc_irq_data = 0;
	spin_unlock_irq(&rtc_lock);

	return 0;
}

static int mr2a_rtc_fasync (int fd, struct file *filp, int on)
{
	return 0;
}

static int mr2a_rtc_release(struct inode *inode, struct file *file)
{
	/*
	 * Turn off any interrupts once the device is no longer
	 * in use, and clear the data.
	 */

	if (file->f_flags & FASYNC) {
		mr2a_rtc_fasync (-1, file, 0);
	}

	MOD_DEC_USE_COUNT;

	atomic_set(&rtc_status, atomic_read(&rtc_status) & ~RTC_IS_OPEN);
	return 0;
}

static unsigned int mr2a_rtc_poll(struct file *file, poll_table *wait)
{
	return 0;
}


/*
 *	The various file operations we support.
 */

static struct file_operations mr2a_rtc_fops = {
	llseek:	mr2a_rtc_llseek,
	read:	mr2a_rtc_read,
	poll:	mr2a_rtc_poll,
	ioctl:	mr2a_rtc_ioctl,
	open:	mr2a_rtc_open,
	release:	mr2a_rtc_release,
	fasync:	mr2a_rtc_fasync,
};



static struct miscdevice mr2a_rtc_dev=
{
	MR2A_RTC_MINOR,
	"mr2artc",
	&mr2a_rtc_fops
};

static int __init mr2a_rtc_init(void)
{
	misc_register(&mr2a_rtc_dev);
	create_proc_read_entry ("driver/mr2artc", 0, 0, mr2a_rtc_read_proc, NULL);

	epoch = 1970;
	epoch_adj = mktime(epoch,0,0,0,0,0) - mktime(1970,0,0,0,0,0);

        udelay(100);
	*(unsigned short *)MR2A_RTC_REGF = 0x0000; /* Jackal modify */

	printk(KERN_INFO "MR-2A Real Time Clock Driver v" MR2A_RTC_VERSION "\n");

	return 0;
}

static void __exit mr2a_rtc_exit (void)
{
	/* interrupts disabled at this point by rtc_release */

	epoch_adj = 0;
	remove_proc_entry ("driver/mr2artc", NULL);
	misc_deregister(&mr2a_rtc_dev);
}

module_init(mr2a_rtc_init);
module_exit(mr2a_rtc_exit);
EXPORT_NO_SYMBOLS;

/*
 *	Info exported via "/proc/driver/mr2artc".
 */

static int rtc_proc_output (char *buf)
{
	char *p;
	struct rtc_time tm;
	
	p = buf;

	get_rtc_time(&tm);

	/*
	 * There is no way to tell if the luser has the RTC set for local
	 * time or for Universal Standard Time (GMT). Probably local though.
	 */
	p += sprintf(p,
		     "mr2a_rtc_time\t: %02d:%02d:%02d\n"
		     "mr2a_rtc_date\t: %04d-%02d-%02d\n"
	 	     "mr2a_rtc_epoch\t: %04lu\n\n",
		     tm.tm_hour, tm.tm_min, tm.tm_sec,
		     tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, epoch);
	return  p - buf;
}

static int mr2a_rtc_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data)
{
	int len = rtc_proc_output (page);
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
		return len;
}

/*
 * Used to disable/enable interrupts for any one of UIE, AIE, PIE.
 * We also clear out any old irq data after an ioctl() that
 * meddles with the interrupt enable/disable bits.
 */
static int set_mr2a_rtc_time(struct rtc_time *rtc_tm)
{
	unsigned int leap_day;
	unsigned int year;
	unsigned long timeval;
        int h;
   
   
	year = rtc_tm->tm_year + 1900;
	leap_day = (rtc_tm->tm_mon == 1) && ((!(year % 4) && (year % 100))
			|| !(year % 400));
		
	if ((year < 2000) || (year > 2099) || (rtc_tm->tm_mon > 11) ||
		(rtc_tm->tm_hour > 24) || (rtc_tm->tm_min > 60) ||
		(rtc_tm->tm_sec > 60) || (rtc_tm->tm_mday == 0) ||
		(rtc_tm->tm_mday > days_in_mo[rtc_tm->tm_mon] + leap_day))
			return -EINVAL;

	spin_lock_irq(&rtc_lock);
	*MR2A_RTC_SEC10 = rtc_tm->tm_sec / 10; 
	*MR2A_RTC_SEC1 = rtc_tm->tm_sec % 10;
	*MR2A_RTC_MIN10 = rtc_tm->tm_min / 10;
	*MR2A_RTC_MIN1 = rtc_tm->tm_min % 10;
   
/* Modify by Jackal */
#if 0   /* 24 hours */
        *MR2A_RTC_HOU10 = rtc_tm->tm_hour / 10;
        *MR2A_RTC_HOU1 = rtc_tm->tm_hour % 10;   
#else   /* 12 hours */
        /* AM */
        if (rtc_tm->tm_hour < 12){
           *MR2A_RTC_HOU10 = 0x0000;
	   h = rtc_tm->tm_hour;
	/* PM */
	}else{
           *MR2A_RTC_HOU10 = 0x0004;
	   h = rtc_tm->tm_hour - 12;
	}
        h = (h == 0) ? 12 : h;
        *MR2A_RTC_HOU10 |= h / 10;
	*MR2A_RTC_HOU1   = h % 10;	   
#endif
	*MR2A_RTC_DAY10 = rtc_tm->tm_mday / 10;
	*MR2A_RTC_DAY1 = rtc_tm->tm_mday % 10;
	*MR2A_RTC_MON10 = (rtc_tm->tm_mon + 1) / 10;
	*MR2A_RTC_MON1 = (rtc_tm->tm_mon + 1) % 10;
	*MR2A_RTC_YEA10 = (rtc_tm->tm_year + 1900 - 2000) / 10;
	*MR2A_RTC_YEA1 = (rtc_tm->tm_year + 1900 - 2000) % 10;
	spin_unlock_irq(&rtc_lock);

	timeval = mktime(year, rtc_tm->tm_mon + 1, rtc_tm->tm_mday,
			 rtc_tm->tm_hour, rtc_tm->tm_min, rtc_tm->tm_sec);
        #if 0 /* kernel version : 2.4.10 */
	set_rtc_time(timeval, VR41XX_ETIMELREG);	// sync to VR41XX_ETIMELREG for users still use VR rtc
        #else /* kernel version : 2.4.18 */
        // rtc_set_time(timeval);
        #endif
     return 0;
}

/*=====================================================================*/
static void get_rtc_time(struct rtc_time *rtc_tm)
{
	/*
	 * Only the values that we read from the RTC are set. We leave
	 * tm_wday, tm_yday and tm_isdst untouched. Even though the
	 * RTC has RTC_DAY_OF_WEEK, we ignore it, as it is only updated
	 * by the RTC when initially set to a non-zero value.
	 */
        int h;
   
	spin_lock_irq(&rtc_lock);
	rtc_tm->tm_sec = *MR2A_RTC_SEC10 * 10 + *MR2A_RTC_SEC1;
	rtc_tm->tm_min = *MR2A_RTC_MIN10 * 10 + *MR2A_RTC_MIN1;
   
/* Modify by Jackal */
#if 0 /* 24 hours */   
	rtc_tm->tm_hour = *MR2A_RTC_HOU10 * 10 + *MR2A_RTC_HOU1;
#else /* 12 hours */
        h = (*(unsigned short *)MR2A_RTC_HOU10 & 0x0001) * 10 + *MR2A_RTC_HOU1;
        /* PM ; (12) 12 - 11 == (24) 12 - 23 */
        if (*MR2A_RTC_HOU10 & 0x0004) {
	   h += 12;
	   h = (h == 24) ? 12 : h;
	/* AM ; (12) 12 - 11 == (24)  0 - 11 */
	}else{
	   h = (h == 12) ? 0 : h;
	}
        rtc_tm->tm_hour = h;
#endif   
	rtc_tm->tm_mday = *MR2A_RTC_DAY10 * 10 + *MR2A_RTC_DAY1;
	rtc_tm->tm_mon = *MR2A_RTC_MON10 * 10 + *MR2A_RTC_MON1 - 1;
	rtc_tm->tm_year = *MR2A_RTC_YEA10 * 10 + *MR2A_RTC_YEA1 + 2000 - 1900;
	spin_unlock_irq(&rtc_lock);
}

unsigned long get_mr2a_rtc_time(void)
{
	struct rtc_time wtime;
	unsigned long timeval;
	
	get_rtc_time(&wtime);

	timeval = mktime(wtime.tm_year + 1900, wtime.tm_mon + 1, wtime.tm_mday,
		wtime.tm_hour, wtime.tm_min, wtime.tm_sec);

	return timeval;
}
/*=====================================================================*/

#ifdef CONFIG_PM
/*
 * Add by Jackal (2002)
 * - For Power Manager
 */
unsigned long pm_mr2artc_offset(void)
{
	unsigned long retval;
	struct rtc_time wtime;
   
	get_rtc_time(&wtime);

	retval = ( (wtime.tm_mday & 0xff) << 24 | (wtime.tm_hour & 0xff) << 16 | 
		   (wtime.tm_min  & 0xff) << 8  | wtime.tm_sec & 0xff );
   
	return retval;
}
#endif
