/*
 *  The driver is PMU(Power Management Unit) Interrupt Control driver.
 *
 *  Now, it supports the folwing function:
 *  While the button of power is pressed, the power is off right now.
 *
 *  Copyright (C) 2001-6-29 CIH <cih@coventive.com>
 *
 *  ChangeLog:
 *	2001-08-03 CIH <cih@coventive.com>
 *		Add low battery interrupt(NMI/Int0) control.
 *	2001-11-07 SAM <sam_liao@coventive.com>
 *		Add ioctl to control PMU
 *      2001-12-14 Tommy
 *              Add etimer.
 *      2002-07-26 Jackal modify.
 *      2002-10-28 Jackal modify.
 *                 New spec as below
 *                 - enter sleep time:
 *                   if time = 0 (min) , stop sleep timer.
 *                   if time > 0 (min) , start sleep timer.
 *                 - backlight to set on/off.
 *                 - backlight off time :
 *                   if time = 0 (min) , stop backlight off timer.
 *                   if time > 0 (min) , start backlight off timer.
 *      2002-12-02 Jackal modify.
 *      2002-12-11 Jackal modify.
 */

#include <linux/config.h>

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/vr4121/vr4121.h>

#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>

#define ACPI_VERSION "2003-06-03 V1.2"
/* 2003-04-04 V1.1 -> 2003-06-03 V1.2:
 * - Add LowPower Status Flags
 * - Fixed calculating time bug
 */

/****************************************************************
 * define 
 ****************************************************************/
/*
 * VR41XX_ACPI_MINOR 154 
 * 
 * How to mknod ? (Jackal comment)
 * - mknod /dev/pms c 154 0
 */
#define PMU_SET_BACKLIGHT           0
#define PMU_SET_LCD                 1 /* setting lcd on/off     */
#define PMU_SET_STANDBY             2 /* standby mode           */
#define PMU_SET_SUSPEND             3 /* suspend mode           */
#define PMU_SET_POWEROFF            4 /* sleep mode / power off */

#define PMU_TIMEOUT_BL             11 /* timer of backlight off */
#define PMU_TIMEOUT_LP             12 /* timer of lowpower mode */
#define PMU_TIMEOUT_SP             13 /* timer of sleep  mode   */

#define PMU_STATUS_CHARGE          21 
#define PMU_STATUS_POWEROFF        22
#define PMU_STATUS_CTRL_POWEROFF   23
#define PMU_SET_STATUS_POWEROFF    24

#define PMU_PWOFF_DISABLE          31
#define PMU_PWOFF_ENABLE           32

#if 0
#define PMU_STATUS_REBOOT          23
#define PMU_SET_REBOOT             51
#endif

#define PMU_SET_SYSOFF             52

#define ON        1
#define OFF       0

/****************************************************************
 * public / private variables
 ****************************************************************/
/*
 * PMU_STATUS : ( Jackal Define )
 * bit  0 -  7 = backlight status 
 * bit  8 - 15 = standby status 
 * bit 16 - 23 = reserve
 * bit 24 - 31 = sleep status
 * 
 * backlight status => bit  0      ( 1 : backlight on , 0 : backlight off );
 *                     bit  1 -  3 ( reserve )
 *                     bit  4      ( 1 : start ; 0 : stop) auto backlight off timer.
 *                     bit  5      ( 1 : runing; 0 ; none)
 *                     bit  6 -  7 ( reserve )
 * lowpower status  => bit  8      ( 1 : start ; 0 : stop) standby mode.
 *                     bit  9      ( 1 : start ; 0 : stop) suspend mode.
 *                     bit 10 - 23 ( reserve )
 * sleep status     => bit 24 - 27 ( reserve )
 *                  => bit 28      ( 1 : start ; 0 : stop) auto power off timer(sleep).
 *                     bit 29 - 31 ( reserve )
 */
#define PMU_STATUS_BL_ON  0x00000001
#define PMU_STATUS_TB     0x00000010
#define PMU_STATUS_TB_RUN 0x00000020

#define PMU_STATUS_TS     0x10000000

#define PMU_STATUS_LO_SD  0x00000100
#define PMU_STATUS_LO_SP  0x00000200
#define PMU_STATUS_LO     (PMU_STATUS_LO_SD | PMU_STATUS_LO_SP)

static unsigned long PMU_STATUS = PMU_STATUS_BL_ON;

static unsigned long PMU_ACTION = 0;

static int PMU_TB_EXPIRES  = 0; 
static int PMU_TS_EXPIRES  = 0;
static int PMU_TB_INPUT    = 0; /* user arg , unit : seconds */
static int PMU_TS_INPUT    = 0; /* user arg , unit : seconds */

static int PMU_TO_EXPIRES  = 2; /* 2 second ; OPTO's spec ...*/

static int VEXTIN_STATUS   = 0; /* 0 : uncharging ; 1 : charging */
static int POWEROFF_STATUS = 0; /* 0 : off ; 1 : on              */

static int CTRL_POWEROFF   = 1; /* 0 : enable poweroff ; >0 : disable poweroff */
static int PMU_TS_INPUT_TMP = 0;

#if 0
static int REBOOT_STATUS   = 0;	
#endif

static unsigned long start_timeflag = 0,end_timeflag = 0;
static struct timer_list pmu_timer;

int battery_low_int = 0;
int buzzflag=0;

/****************************************************************
 * function prototype 
 ****************************************************************/
static int pm_time_offset(unsigned long,unsigned long);

       void pmu_set_backlight(int action);
       void pmu_set_lcd(int action);
static void pmu_set_standby(void);
static void pmu_set_suspend(void);
       void pmu_set_poweroff(void);

#ifdef CONFIG_PM_SLEEP
static void pmu_set_sleep(void);
#endif

       void pmu_timer_lowpower(int action);
static void pmu_set_interval(unsigned long pmustatus, int interval);
static void pmu_timer_fun(void);
static void pmu_timer_start(void);
static void pmu_timer_stop(void);
static void pmu_set_bl_status(int action);
static void pmu_timer_action(void);

        int pm_access_device(void);

/* proc function */
static int pmu_proc_output (char *buf);
static int pmu_status_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data);
/* ioctl function */
static int pmu_open(struct inode *inode, struct file *filp);
static int pmu_close(struct inode *inode, struct file *filp);
static int pmu_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);

static void pmu_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs);
static void battery_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs);
int __init vr41xx_acpi_init(void);

/* extern */
extern void ctrl_alt_del(void);

/* arch/mips/vr41xx/common/power.c */
#ifdef CONFIG_PM_STANDBY
extern void PM_do_standby(void);
#endif
#ifdef CONFIG_PM_POWERED_SUSPEND
extern void PM_do_suspend(void);
#endif
#ifdef CONFIG_PM_SLEEP
extern int PM_do_sleep(void);
#endif

#ifdef CONFIG_BUZZER
extern void low_power_buzz(void);
extern void keyboard_irq_buzz(void);
#endif

extern unsigned long pm_mr2artc_offset(void);

/*
 * Generic
 */
spinlock_t pmu_etimer_lock = SPIN_LOCK_UNLOCKED;
        
static inline unsigned short read_etime_register(volatile unsigned short *addr)
{
   unsigned short val;
   
   do{
      val = *addr;
   }while (val != *addr);
   
   return val;
}

static unsigned long etimer_get_time(void)
{
   unsigned short etimel, etimem, etimeh;
   
   do{
      etimem = read_etime_register(VR41XX_ETIMEMREG);
      etimeh = read_etime_register(VR41XX_ETIMEHREG);
      etimel = read_etime_register(VR41XX_ETIMELREG);
   }while (etimem != read_etime_register(VR41XX_ETIMEMREG));
    
   return ((etimeh << 17) | (etimem << 1) | (etimel >> 15));
}

static int etimer_set_compare_time(unsigned long sec)
{
   unsigned long flags;

   spin_lock_irqsave(&pmu_etimer_lock, flags);
   *VR41XX_ECMPLREG = sec << 15;
   *VR41XX_ECMPMREG = sec >> 1;
   *VR41XX_ECMPHREG = sec >> 17;
   spin_unlock_irqrestore(&pmu_etimer_ock, flags);
   
   return 0;
}

static void pmu_etimer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
   unsigned short val;
   
   val = *VR41XX_RTCINTREG;
   barrier();
   *VR41XX_RTCINTREG = val;
}

static int pmu_etimer_setup(unsigned long sec)
{
   //unsigned short val;
   int ret;

   disable_irq(VR41XX_IRQ_ETIMER);
   
   /* ack int */
   *VR41XX_RTCINTREG = 0x0001;
   barrier();
   
   sec += etimer_get_time();
   etimer_set_compare_time(sec);
   
   ret = request_irq(VR41XX_IRQ_ETIMER,pmu_etimer_interrupt,0,"etimer",NULL);
   if (ret){
      printk(KERN_INFO "pmu_etimer_setup : can't get assigned irq %i\n", VR41XX_IRQ_ETIMER );
      return 1;
   }
   
   //enable_irq(VR41XX_IRQ_ETIMER);
   return 0;
}

static void pmu_etimer_free(void)
{
   disable_irq(VR41XX_IRQ_ETIMER);
   free_irq(VR41XX_IRQ_ETIMER,NULL);
   barrier();
}

/*
 * max : 36400 (sec)
 */
static int pm_time_offset(unsigned long starttime,unsigned long endtime)
{
   int h1,m1,s1;
   int h2,m2,s2;
   
   h1 = (int)((starttime >> 16) & 0xff);
   m1 = (int)((starttime >>  8) & 0xff);
   s1 = (int)( starttime & 0xff);
   
   h2 = (int)((endtime >> 16) & 0xff);
   m2 = (int)((endtime >>  8) & 0xff);
   s2 = (int)( endtime & 0xff);

   
   /* seconds */
   s1 = h1*60*60 + m1*60 + s1;
   s2 = h2*60*60 + m2*60 + s2;

   /* default */
   h1 = 24*60*60;
   
   if (s2 < s1) 
     h2 = h1 - s1 + s2;
   else
     h2 = s2 - s1;
   
   return h2;
}


void pmu_timer_lowpower(int action)
{
   int timeflag=0;
   int k;
   /* 
    * define : flag
    * 
    * 0: nothing
    * 1: lowpower_time < pm_time;
    * 2: lowpower_time > pm_time;
    * 3: lowpower_time = pm_time;
    */
   int flag; 

   if (CTRL_POWEROFF) return;
   
   flag = 0;
   
   
   if (PMU_ACTION){
      pmu_timer_stop();
      
      /* setting a time flag of ending */
      end_timeflag = pm_mr2artc_offset();
      
      timeflag = pm_time_offset(start_timeflag,end_timeflag);
      
      if (timeflag >= 0){
	 switch(PMU_ACTION){
	  case PMU_STATUS_TB:
	    k = PMU_TB_EXPIRES/HZ;
	    if (timeflag <= k)
	      timeflag = k - timeflag;
	    else
	      timeflag = -1;
	    break;
	  case PMU_STATUS_TS:
	    k = PMU_TS_EXPIRES/HZ;
	    if (timeflag <= k)
	      timeflag = PMU_TS_EXPIRES/HZ - timeflag;
	    else
	      timeflag = -1;
	    break;
	 }
	 
	 if (timeflag < PMU_TO_EXPIRES)
	   flag = 1;
	 else if (timeflag > PMU_TO_EXPIRES)
	   flag = 2;
	 else if (timeflag == PMU_TO_EXPIRES)
	   flag = 3;
      }
   }

   //printk("@@@@@ flag = %d,timeflag = %d\n",flag, timeflag);
   buzzflag = 1;
   pmu_etimer_setup(PMU_TO_EXPIRES);
   //printk("@@@@@ setup etimer\n");
   
   if (!POWEROFF_STATUS){
      switch(action){
       case 1:
	 PMU_STATUS |=  PMU_STATUS_LO_SD;
	 pmu_set_standby();
	 PMU_STATUS &= ~PMU_STATUS_LO_SD;	 
	 break;
       case 2:
	 PMU_STATUS |=  PMU_STATUS_LO_SP;	 
	 pmu_set_suspend();
	 PMU_STATUS &= ~PMU_STATUS_LO_SP;
	 break;
      }
   }
   
   pmu_etimer_free();   
   buzzflag = 0;
   
   switch(flag){
    case 2:
      switch(PMU_ACTION){
       case PMU_STATUS_TB:
	 PMU_TB_EXPIRES = (timeflag - PMU_TO_EXPIRES)*HZ;
	 break;
       case PMU_STATUS_TS:
	 PMU_TS_EXPIRES = (timeflag - PMU_TO_EXPIRES)*HZ;
	 break;
      }
      pmu_timer_start();
      break;
    case 1:
    case 3:
      pmu_timer_fun();
      break;
   }
}
/****************************************************************
 * START : PM (Power Manager) standard function
 ****************************************************************/
static void pmu_control_pwoff(int action)
{
   switch(action){
    case 1:
      if (CTRL_POWEROFF <= 0)
	CTRL_POWEROFF = 0;
      else
	CTRL_POWEROFF--;
      break;
    case 0:
      if (CTRL_POWEROFF >= 0xFFFFFFFF)
	CTRL_POWEROFF = 0xFFFFFFFF;
      else
	CTRL_POWEROFF++;
      break;
   }
}

void pmu_disable_pwoff(void)
{
   pmu_control_pwoff(0);
}

void pmu_enable_pwoff(void)
{
   pmu_control_pwoff(1);
}

int pmu_pwoff_status(void)
{
   return CTRL_POWEROFF;
}

/*
 * backlight on/off
 */
void pmu_set_backlight(int action)
{
   /* MR-2 GPIO6 set writeable */
   *(unsigned short *)(0x0a000040 + KSEG1) |= 0x0040;

   if (action)
      /* MR-2 GPIO6 high; back light on */     
      *(unsigned short *)(0x0a000044 + KSEG1) |= 0x0040;
   else
      /* MR-2 GPIO6 low ; back light off */
      *(unsigned short *)(0x0a000044 + KSEG1) &= ~0x0040;
}

/*
 * lcd on/off
 */
void pmu_set_lcd(int action)
{
   /* MR-2 GPIO5 set writeable */
   *(unsigned short *)(0x0a000040 + KSEG1) |= 0x0020;
   
   if (action)
     /* MR-2 GPIO5 high; back light on */     
     *(unsigned short *)(0x0a000044 + KSEG1) |= 0x0020;
   else
     /* MR-2 GPIO5 low ; back light off */     
     *(unsigned short *)(0x0a000044 + KSEG1) &= ~0x0020;
}

/*
 * set standby mode
 */
static void pmu_set_standby(void)
{
   #ifdef CONFIG_PM_STANDBY
   PM_do_standby();
   #else
   printk("Not Define \"statndby\" ...\n");
   #endif
}

static void pmu_set_suspend(void)
{
   #ifdef CONFIG_PM_POWERED_SUSPEND
   PM_do_suspend();
   //printk("@@@@@ pmu_set_suspend function\n");
   #else
   printk("Not Define \"suspend\" ...\n");
   #endif
}

/*
 * power off
 */
void pmu_set_poweroff(void)
{
   
   printk(KERN_INFO"Will turn off your system now!!!\n");

#ifdef CONFIG_BUZZER
   low_power_buzz();
#endif
   mdelay(100);
   
   /* 
    * jackal add 20020917 
    *   - MR2-A register of "PDECGPIOSEL" 
    */
   *(unsigned short *)(0xaa00005c) &= 0x0000;
   
   *((unsigned short *)0xab000104) &= 0x00F0; /* VR4121-GIUPIODL- GPIO Level Low */
   *((unsigned short *)0xaa000044) &= 0x0000; /* MR-2A-GIUPIOD-GPIO Level All Low */
   *((unsigned short *)0xaa000260) &= 0x8090; /* MR-2A-CMUCLKMSK-Clock Mask Setting */
   *((unsigned short *)0xaa000262) &= 0x582F; /* MR-2A-CMUSRST-Software Reset ON */
   
   *((unsigned short *)0xaa000282) |= 0x0001; /* cpu_off_cmd_reg */
   *((unsigned short *)0xaa000282) |= 0x0001; /* cpu_off_cmd_reg */
}

#ifdef CONFIG_PM_SLEEP
/*
 * set sleep mode
 */
static void pmu_set_sleep(void)
{
   POWEROFF_STATUS = 2;
   mdelay(100);
   
   PM_do_sleep();
   
   POWEROFF_STATUS = 0;
}
#endif

/****************************************************************
 * START : PM (Power Manager) timer 
 ****************************************************************/
static void pmu_set_interval(unsigned long pmustatus, int interval)
{
   if (interval >= 0 ) {
      switch(pmustatus){
       case PMU_STATUS_TB:
	 PMU_TB_INPUT = interval;
	 
	 if (!PMU_TB_INPUT){
	    if (PMU_STATUS & PMU_STATUS_BL_ON)
	      pmu_set_backlight(ON);
	    else
	      pmu_set_backlight(OFF);
	 }
	 break;
       case PMU_STATUS_TS:
	 PMU_TS_INPUT = interval;
	 PMU_TS_INPUT_TMP = interval;
	 break;
      }
   }
}

static void pmu_timer_fun(void)
{
   if (PMU_STATUS & PMU_ACTION){
      switch(PMU_ACTION){
       case PMU_STATUS_TB:
	 pmu_set_backlight(OFF);
	 
	 if (PMU_STATUS & PMU_STATUS_TS){
	    PMU_ACTION = PMU_STATUS_TS;
	    mod_timer(&pmu_timer, jiffies + PMU_TS_EXPIRES);
	 }
	 break;
       case PMU_STATUS_TS:
	 POWEROFF_STATUS = 1;
	 break;
      }
   }else
     PMU_ACTION = 0;
}

static void pmu_timer_start(void)
{
   switch(PMU_ACTION){
      
    case PMU_STATUS_TB:
      init_timer(&pmu_timer);
      pmu_timer.function = (void (*)(unsigned long))pmu_timer_fun;
      pmu_timer.expires  = jiffies + PMU_TB_EXPIRES;
      
      /* setting a start flag of timer */
      start_timeflag=pm_mr2artc_offset();
      
      add_timer(&pmu_timer);
      break;
      
    case PMU_STATUS_TS:
      init_timer(&pmu_timer);
      pmu_timer.function = (void (*)(unsigned long))pmu_timer_fun;
      pmu_timer.expires  = jiffies + PMU_TS_EXPIRES;
      
      /* setting a start flag of timer */
      start_timeflag=pm_mr2artc_offset();
      
      add_timer(&pmu_timer);
      break;
   }
}
static void pmu_timer_stop(void)
{
   del_timer(&pmu_timer);
}

static void pmu_set_bl_status(int action)
{
   if (action)
     PMU_STATUS |=  PMU_STATUS_BL_ON;
   else
     PMU_STATUS &= ~PMU_STATUS_BL_ON;
}

static void pmu_timer_action(void)
{
   int tmp = 0, flag = 0;
   
   if ( !(PMU_STATUS & PMU_STATUS_BL_ON) ){
      flag = 1;
      tmp  = PMU_TB_INPUT;
      PMU_TB_INPUT = 0;
   }
   
   /*
    * sleep time         : TS
    * backlight off time : TB
    * 
    * case 1: TB -> TS 
    *       ( TS &&  TB) && (TS >  TB)
    * case 2: TS
    *       ( TS && !TB) || (TS <= TB && TS)
    * case 3: TB
    *       (!TS &&  TB)
    * case 4: NONE
    *       (!TS && !TB)
    */
   
   /* case 1: TB -> TS */
   if ((PMU_TS_INPUT && PMU_TB_INPUT) && (PMU_TS_INPUT > PMU_TB_INPUT)) {
      PMU_TB_EXPIRES =  PMU_TB_INPUT*HZ;
      PMU_TS_EXPIRES = (PMU_TS_INPUT - PMU_TB_INPUT)*HZ;
      
      PMU_STATUS |= (PMU_STATUS_TB | PMU_STATUS_TS);
      
      PMU_ACTION  =  PMU_STATUS_TB;
   /* case 2: TS */
   }else if ((PMU_TS_INPUT && !PMU_TB_INPUT) || (PMU_TS_INPUT <= PMU_TB_INPUT && PMU_TS_INPUT))	{
      PMU_TS_EXPIRES = PMU_TS_INPUT*HZ;
      PMU_TB_EXPIRES = 0;
      
      PMU_STATUS &= ~PMU_STATUS_TB;
      PMU_STATUS |=  PMU_STATUS_TS;
      
      PMU_ACTION  = PMU_STATUS_TS;
   /* case 3: TB */
   }else if (!PMU_TS_INPUT && PMU_TB_INPUT){
      PMU_TB_EXPIRES = PMU_TB_INPUT*HZ;
      PMU_TS_EXPIRES = 0;
      
      PMU_STATUS &= ~PMU_STATUS_TS;
      PMU_STATUS |=  PMU_STATUS_TB;
      
      PMU_ACTION  =  PMU_STATUS_TB;
   /* case 4: NONE */
   }else if ( (!PMU_TS_INPUT) && (!PMU_TB_INPUT) ){
      PMU_TB_EXPIRES = 0;
      PMU_TS_EXPIRES = 0;
      
      PMU_STATUS &= ~(PMU_STATUS_TB|PMU_STATUS_TS);
      
      PMU_ACTION  = 0;
   }
   
   if (flag) PMU_TB_INPUT = tmp;
   
   pmu_timer_stop();
   
   if (PMU_ACTION) 
     pmu_timer_start();
}

int pm_access_device(void)
{
   int retval = 0;
   static unsigned long  start_tm = 0;
   unsigned long end_tm;
   int tm;

   
   end_tm = pm_mr2artc_offset();
   tm = pm_time_offset(start_tm,end_tm);
   start_tm = end_tm;
   
   //printk(__FILE__":"__FUNCTION__"tm=%d\n",tm);
   
   if (!tm) return 0;
   
   if (PMU_ACTION){
      if ( (PMU_STATUS & PMU_STATUS_BL_ON))
	pmu_set_backlight(ON);
      else
	pmu_set_backlight(OFF);

      if ( !(PMU_STATUS & PMU_STATUS_LO) ){
	 pmu_timer_action();
      }
   }
   
   return retval;
}

/**************************************************************
 * START : /proc/driver/pms
 **************************************************************/
static int pmu_proc_output (char *buf)
{
   char *p;

   p = buf;

   p += sprintf(p,"Power Management System\n"
		"backlight timeout : %d (min)\n"
		"power off timeout : %d (min)\n\n",
		PMU_TB_INPUT,PMU_TS_INPUT);
      
   return  p - buf;
}


static int pmu_status_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{
   int len = pmu_proc_output (page);
   if (len <= off+count) *eof = 1;
   *start = page + off;
   len -= off;
   if (len>count) len = count;
   if (len<0) len = 0;
   return len;
}

/**************************************************************
 * START : ioctl function
 **************************************************************/

static int pmu_open(struct inode *inode, struct file *filp)
{
    return 0;
}
static int pmu_close(struct inode *inode, struct file *filp)
{
    return 0;
}

static int pmu_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
   int intval,flag = 0;

   switch(cmd){
    /* (1) setting backlight on/off */
    case PMU_SET_BACKLIGHT:
      if (get_user(intval, (int *)(arg))) return -EFAULT;
      if (intval == ON || intval == OFF) {
	 pmu_set_backlight(intval);
	 pmu_set_bl_status(intval);
	 flag = 1;
      }
      break;
      /* (2) setting lcd on/off */      
    case PMU_SET_LCD:
      if (get_user(intval, (int *)(arg))) return -EFAULT;
      if (intval == ON || intval == OFF)
	pmu_set_lcd(intval);
      break;
      /* (3) setting standby mode */
    case PMU_SET_STANDBY:
      #ifdef CONFIG_PM_STANDBY
      pmu_timer_lowpower(1);
      #else
      printk("Not Define \"standby\" !!!\n");
      #endif
      break;
      /* (4) setting suspend mode */
    case PMU_SET_SUSPEND:
      #ifdef CONFIG_PM_POWERED_SUSPEND
      pmu_timer_lowpower(2);     
      //printk("@@@@@ pmu.c -> 752\n");
      #else
      printk("Not Define \"suspend\" !!!\n");
      #endif
      break;
      /* (5) setting sleep mode / power off */
    case PMU_SET_POWEROFF:
      #ifdef CONFIG_PM_SLEEP
      pmu_set_sleep();
      #else
      pmu_set_poweroff();
      #endif      
      break;
      /* (6) power switch enable/disable*/
    case PMU_PWOFF_DISABLE:
      pmu_disable_pwoff();
      flag = 1;
      break;
    case PMU_PWOFF_ENABLE:
      pmu_enable_pwoff();
      flag = 1;
      break;
      /* power off */
    case PMU_SET_SYSOFF:
      pmu_set_poweroff();      
      break;
#if 0      
      /* reboot */ 
    case PMU_SET_REBOOT:
      vr41xx_restart(0);
      break;
#endif     
      
    /***********************************************************
     * About PM ( Power Manager )
     *  - BL : backlight off time.
     *  - PO : enter sleep time.
     */
    case PMU_TIMEOUT_BL:
      if (get_user(intval, (int *)(arg))) return -EFAULT;
      if (intval < 0) return  -EFAULT;
      flag = 1;
      pmu_set_interval(PMU_STATUS_TB,intval);
      break;
    case PMU_TIMEOUT_SP:
      if (get_user(intval, (int *)(arg))) return -EFAULT;
      if (intval < 0) return  -EFAULT;
      flag = 1;
      pmu_set_interval(PMU_STATUS_TS,intval);
      break;
    /*************************************************************/
      
    case PMU_STATUS_CHARGE:
      VEXTIN_STATUS = (*(unsigned short *)0xaa000280 >> 1) & 1;
      if (copy_to_user((void *)arg, &VEXTIN_STATUS , sizeof(VEXTIN_STATUS)))
	 return -EFAULT;
      break;
    case PMU_STATUS_POWEROFF:
      if (copy_to_user((void *)arg, &POWEROFF_STATUS , sizeof(POWEROFF_STATUS)))
	return -EFAULT;
      break;
    case PMU_SET_STATUS_POWEROFF:
      if (get_user(intval, (int *)(arg))) 
        return -EFAULT;
      POWEROFF_STATUS = intval;
      break;
    case PMU_STATUS_CTRL_POWEROFF:
      intval=(CTRL_POWEROFF) ? 0 : 1;
      if (copy_to_user((void *)arg, &intval , sizeof(intval)))
	return -EFAULT;
      break;
#if 0      
    case PMU_STATUS_REBOOT:
      if (copy_to_user((void *)arg, &REBOOT_STATUS , sizeof(REBOOT_STATUS)))
        return -EFAULT;
      REBOOT_STATUS = 0;
      break;
#endif      
   }
	
   if (flag){
      if (CTRL_POWEROFF){
	 PMU_TS_INPUT = 0;
      }else{
	 PMU_TS_INPUT = PMU_TS_INPUT_TMP;
      }
      pmu_timer_action();
   }
   
   return 0;
}

static struct file_operations pmu_control_fops = {
	ioctl:		pmu_ioctl,
	open:		pmu_open,
	release:	pmu_close,
};

static struct miscdevice vr41xx_acpi_dev = {
   VR41XX_ACPI_MINOR,"vr41xx_pmu", &pmu_control_fops
};
/**************************************************************
 * END : ioctl function.
 **************************************************************/

static void pmu_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs)
{
   unsigned short status;
   unsigned short i;
   
   status = *((unsigned short *)0xaa0002b0);
   
   /* MR-2A PMU_INT_ST_CLR */
   *((unsigned short *)0xaa0002b0) &= ~0x0003;
   /*Clear VR4121 POWERSWINTR status */
   *VR41XX_PMUINTREG |= 0x0001;
      
   /* VEXTIN interrupt */
   if(status & 0x0002){
      VEXTIN_STATUS = (*((unsigned short *)(0xaa000280)) & 0x0002)>>1;
   }

   if (!CTRL_POWEROFF){
      
      /* PWSW interrupt */
      if(status & 0x0001){
	 /* Jeff add */
	 cli();      
	 /* GPIO30, PW pin, input */
	 *( (volatile unsigned short *)0xab000102 ) &= ~0x4000;
	 
	 /* Check 300 ms Low status */
	 for(i = 0; i < 6; i ++) {
	    mdelay(30);
	    if(*((volatile unsigned short *)0xab000106) & 0x4000)
	      {;}
	    else 
	      {
		 /* In-valid, GPIO30 to be interrupt pin */
		 *(unsigned short *)( VR41XX_GIUIOSELH )    &= ~0x4000;
		 *(unsigned short *)( VR41XX_GIUINTTYPH )   |= 0x4000 ;
		 *(unsigned short *)( VR41XX_GIUINTENH )    |= 0x4000 ;
		 *(unsigned short *)( VR41XX_GIUINTHTSELH ) |= 0x4000 ;
		 *(unsigned short *)( VR41XX_GIUINTSTATH )  |= 0x4000 ;
		 sti();
		 
		 return;
	      }
	 }
	 
         while(*((volatile unsigned short *)0xab000106) & 0x4000);
	 //buzz_poweroff_init();
	 
#ifdef CONFIG_PM_SLEEP
	 POWEROFF_STATUS = 1;
#else      
	 pmu_set_poweroff();
#endif
	 sti();
      }
   }
}

static void battery_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs)
{
   unsigned short status;
   int flag =0;
   extern void vr41xx_restart(char *c);

   status = *(unsigned short *)0xaa0002ac;

   // MR-2A NMI_ST_CLR
   *((unsigned short *)0xaa0002ac) &= ~0x0007;
   // Clear VR4121 BATTINTR status
   *VR41XX_PMUINTREG |= 0x0002;
   
   /* PF_LI interrupt */
   if(status & 0x0001){
      printk(KERN_INFO"Battery Low !!!\n");
      battery_low_int = 1;
      flag=1;
   }
   /* BATTSW interrupt */
   if(status & 0x0002){
      printk(KERN_INFO"Battery cover open !!!\n");
      flag=1;
   }
   
   /* RESETSW interrupt */
   if(status & 0x0004){
      *((unsigned short *)0xaa0002a0) = 0;
      
      /* ORANGE */
      *((unsigned short*)(0xab000100)) |= 0x3000;
      *((unsigned short*)(0xab000104)) |= 0x3000;
      flag = 0;
      ctrl_alt_del();	
   }
   
   if (flag){
#ifdef CONFIG_PM_SLEEP
      POWEROFF_STATUS = 1;
#else
      pmu_set_poweroff();
#endif      
   }
}

/*
 * driver/char/misc.c (fun : misc_init)
 */ 
int __init vr41xx_acpi_init(void)
{
   int err;
   
   printk(KERN_INFO"VR41XX ACPI initial .v%s\n",ACPI_VERSION);
   
   /*********************************************************************/
   // irq = 9
   request_irq(VR41XX_IRQ_POWER, pmu_interrupt_handler,0, "power", NULL);   
   
   // (MR2A) ack interrupt
   *((unsigned short *)0xaa0002b0) &= ~0x0003;
   // (MR2A) mask interrupt 
   *((unsigned short *)0xaa0002b2) |= 0x0003;

   // irq = 8 
   request_irq(VR41XX_IRQ_BATTERY, battery_interrupt_handler, 0, "battery", NULL);
   // (MR2A) ack battery interrupt 
   *((unsigned short *)0xaa0002ac) &= ~0x0007;
   // MR-2A  mask battery interrupt
   *((unsigned short *)0xaa0002ae) |= 0x0007; 
   
   // (MR2A) PF_LI/PF_AM3 Signal Sampling Time Set Register
   *((unsigned short *)0xaa00028a) |= 0x000f;
   
   // (MR2A) NMI forced off timer clear
   *((unsigned short *)0xaa0002a0) &= 0xfffe;
   
   // (VR4121) ack battint 
   *VR41XX_PMUINTREG |= 0x0002;
   
   // Low battery detect interrupt type is NMI
   // VR4121 NMIREG
   *VR41XX_NMIREG |= 0x0001; // Int0
   
   // VR4121 battery interrupt enable
   // VR4121 MSYSINT1REG
   *VR41XX_MSYSINT1REG |= 0x0001;
   /*********************************************************************/
   err = misc_register(&vr41xx_acpi_dev);
   
   create_proc_read_entry ("driver/pms", 0, 0, pmu_status_proc, NULL);
   
   if (err < 0) {
      printk(KERN_WARNING "ACPI Controller Misc Register err %d\n",err);
      return err;
   }
   
   return 0;
}
