/*
 * linux/arch/mips/vr41xx/roadht/prom.c
 *
 * Copyright (C) 1999 Bradley D. LaRonde and Michael Klar
 *
 * Copyright (C) 2001 Jalen Yang <jalen@coventive.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/bootmem.h>
#include <asm/bootinfo.h>
#include <asm/addrspace.h>
#include <asm/page.h>


char arcs_cmdline[CL_SIZE];
extern int _end;

//#define min(a,b)  (((a)<(b))?(a):(b))

#ifndef CONFIG_CMDLINE
#define CONFIG_CMDLINE ""
#endif

const char *get_system_type(void)
{
	        return "ROAD HT ";
}

unsigned long __init probe_ram_size(void)
{
	// this figures out the actual amount of RAM present, in increments of 1MB, max 64MB
	// returns the first physical address just past the end of RAM -
	// IOW the amount of ram detected

	unsigned long ramsize;
	unsigned int * p1, * p2;
	unsigned int d1, d2;

	// begin probe starting with the first 1MB past the global data area
	ramsize = ((unsigned long)&_end + 0x000fffff) & 0x1ff00000;

	// probe sequential 1MB areas, looking for a lack of RAM
	while (ramsize < 0x04000000) {
		// save val at a spot  and replace it with a pattern
		p1 = (unsigned int *)(KSEG1 + ramsize + 0x000ffff0);
		d1 = *p1;
		*p1 = 0x1234a5a5;
		barrier();

		// save a val at a second spot and replace with another pattern
		// this second write thwarts an empty bus faking us out
		p2 = (unsigned int *)(KSEG1 + ramsize + 0x000fffe0);
		d2 = *p2;
		*p2 = 0x00ff00ff;
		barrier();

		// isn't the first pattern still there?
		if (*p1 != 0x1234a5a5)
			// we are past the end of ram, so quit
			break;

		// restore what was there
		*p1 = d1;
		*p2 = d2;

		// check next 1MB area
		ramsize += 0x00100000;
	}

	// return the amount of ram detected
	return ramsize; 
}

int __init prom_init(int argc, char **argv, char **envp)
{
	unsigned long mem_limit, mem_detected, mem_use;
	unsigned long mem_start, mem_end, bootmap_size;
	int i;

	// clear ERL and EXL in case the bootloader got us here through an
	// exception.  this is mostly to make kernel debugger happy
	write_32bit_cp0_register(CP0_STATUS, 0);

	// set upper limit to vr41xx maximum physical RAM (64MB)
	mem_limit = 64 << 20;

	// the bootloader passes us argc/argv[], but the kernel wants one big string
	// put it in arcs_cmdline, which later gets copied to command_line
	// (see arch/mips/kernel/setup.c)
	// we skip the first argument which is the program file name
	strcpy(arcs_cmdline, "");
	for (i = 1; i < argc; i++) {
		// is it the "mem=" option?
		if ( memcmp(argv[i], "mem=", 4) == 0 )
		{
			// limit to amount of memory specified in megabytes
			mem_limit = simple_strtoul(argv[i] + 4, 0, 10) << 20;

			printk("Command line argument limits memory to %dMB.\n",
				(int)mem_limit >> 20);

			// don't pass this arg on to the kernel
			continue;
		}

		if ( strlen(arcs_cmdline) > 0 )
			strcat(arcs_cmdline, " ");
		strcat(arcs_cmdline, argv[i]);
	}

	strcat(arcs_cmdline, CONFIG_CMDLINE);

	mips_machgroup = MACH_GROUP_NEC_VR41XX;
	mips_machtype = MACH_VR41XX_ROAD_HT;

	// probe for the amount of ram present
	// TODO: boot param "noramprobe" inhibits this probe; must use with boot param "ram"
	mem_detected = probe_ram_size();

	// don't use more RAM than we detected
	mem_use = min(mem_limit, mem_detected);

	printk("Detected %dMB of memory.  Will use %dMB of it.\n",
		(int)mem_detected >> 20, (int)mem_use >> 20);

	// TODO: rework this so it allows for memory holes.  some devices have them.

	mem_start = __pa(PAGE_ALIGN((unsigned long)&_end));
	mem_end = mem_use;

	// Register all the contiguous memory with the bootmem allocator
	// and free it.  Be careful about the bootmem freemap.
	bootmap_size = init_bootmem(mem_start >> PAGE_SHIFT, mem_end >> PAGE_SHIFT);
	free_bootmem(mem_start + bootmap_size, mem_end - mem_start - bootmap_size);

	return 0;
}

int __init page_is_ram(unsigned long pagenr)
{
	return 1;
}

void __init prom_fixup_mem_map(unsigned long start, unsigned long end)
{
}

void prom_free_prom_memory(void)
{
}

